@@ -6,6 +6,28 @@ from pathlib import Path
 
 # An extension of the music21 note class with more information easily accessible
 class NoteListElement:
+    """
+    An extension of the music21 note class
+
+    Attributes
+    ----------
+    note : m21.note.Note
+        music21 note class
+    offset : int
+        cumulative offset of note
+    id : int
+        unique music21 id
+    metadata : music21.metadata
+        piece metadata- not normally attached to a music21 note
+    part : str
+        voice name
+    partNumber : int
+        voice number, not 0 indexed
+    duration : int
+        note duration
+    piece_url : str
+        piece url for note
+    """
     def __init__(self, note: m21.note.Note, metadata, part, partNumber, duration, piece_url):
         self.note = note
         self.offset = self.note.offset
@@ -22,8 +44,40 @@ class NoteListElement:
 # For mass file uploads, only compatible for whole piece analysis, more specific tuning to come
 class CorpusBase:
     # Need to consider whether users can input certain scores (which means needing urls selected too), or just to do all in the corpus automatically
+    """
+    A class for importing multiple scores at once
+
+    Attributes
+    ----------
+    urls : list
+        list of urls of scores to be imported
+    paths : list
+        list of file paths of scores to be imported
+    scores : list of music21.Score
+        list of music21.Score objects- imported from urls and paths
+    all_urls : list
+        stores both paths and urls- not for user usage
+    note_list : list of NoteListElement
+        list of notes constructed from scores
+    note_list_no_unisons : list of NoteListElement
+        list of notes constructed from scores, combining unisons
+    """
     def __init__(self, urls:list, paths:list):
+        """
+        Parameters
+        ----------
+        urls : list
+            list urls to mei files
+        paths : list
+            list paths to mei files
+
+        Raises
+        ----------
+        Exception
+            If at least one score isn't succesfully imported, raises error
+        """
         self.urls = urls
+        self.paths = paths
         self.scores = []
         self.all_urls = []
         for url in urls:
@@ -49,6 +103,8 @@ class CorpusBase:
         self.no_unisons = self.note_list_no_unisons()
 
     def note_list_whole_piece(self):
+        """ Creates a note list from the whole piece for all scores- default note_list
+        """
         pure_notes = []
         urls_index = 0
         for score in self.scores:
@@ -71,6 +127,11 @@ class CorpusBase:
         return pure_notes
 
     def note_list_no_unisons(self):
+        """ Creates a note list from the whole piece for all scores combining unisons
+
+        Combines consecutive notes at the same pitch into one note, adding in the duration
+        of the next note into the previous one.
+        """
         pure_notes = []
         urls_index = 0
         for score in self.scores:
@@ -94,6 +155,14 @@ class CorpusBase:
         return pure_notes
 
     def note_list_selected_offset(self, offsets: list):
+        """
+        Creates a note list from the whole piece for all scores, going by provided offsets
+
+        Parameters
+        ----------
+        offsets : list
+            offsets within measures to collect notes at (notes collected will be those that are sounding at that offset- not just starting)
+        """
         pure_notes = []
         urls_index = 0
         for score in self.scores:
@@ -115,7 +184,29 @@ class CorpusBase:
 
 # For single file uploads
 class ScoreBase:
+    """
+    A class for importing a single score- offers more precise construction options
+
+    Attributes
+    ----------
+    url : str
+        url or path of mei file
+    score : music21.Score
+        music21.Score object gathered from mei file import
+    note_list : list of NoteListElement
+        list of notes constructed from score
+    """
     def __init__(self, url):
+        """
+        Parameters
+        ----------
+        url:
+            url or path of mei file
+        Raises
+        ----------
+        Exception
+            If score isn't succesfully imported, raises error
+        """
         self.url = url
         print("Requesting file from " + str(self.url) + "...")
         # Detect if local file of url based on leading /
@@ -132,9 +223,10 @@ class ScoreBase:
             except:
                 raise Exception("Import from " + str(self.url) + " failed, please check your url/file type")
         self.note_list = self.note_list_whole_piece()
-        self.note_list_down_beats = self.note_list_selected_beats([1,3])
 
     def note_list_whole_piece(self):
+        """ Creates a note list from the whole piece- default note_list
+        """
         pure_notes = []
         parts = self.score.getElementsByClass(stream.Part)
         for part in parts:
@@ -155,6 +247,11 @@ class ScoreBase:
 
     # Combines unison intervals into one note- generally for increased pattern finding
     def note_list_no_unisons(self):
+        """ Creates a note list from the whole piece for all scores combining unisons
+
+        Combines consecutive notes at the same pitch into one note, adding in the duration
+        of the next note into the previous one.
+        """
         pure_notes = []
         urls_index = 0
         parts = self.score.getElementsByClass(stream.Part)
@@ -178,6 +275,14 @@ class ScoreBase:
 
     # Gets only notes that start on the specified beats- allows for user specification in case of weird time signatures
     def note_list_selected_beats(self, beats: list):
+        """
+        Creates a note list from the whole piece, going by provided beats
+
+        Parameters
+        ----------
+        beats : list
+            collects all notes which begin on specified beat
+        """
         pure_notes = []
         parts = self.score.getElementsByClass(stream.Part)
         for part in parts:
@@ -190,8 +295,15 @@ class ScoreBase:
             pure_notes.append(note_obj)
         return pure_notes
 
-    # Better tuned for different time signatures
     def note_list_by_offset(self, offsets:list):
+        """
+        Creates a note list from the whole piece, going by provided offsets
+
+        Parameters
+        ----------
+        offsets : list
+            offsets within measures to collect notes at (notes collected will be those that are sounding at that offset- not just starting)
+        """
         pure_notes = []
         part_number = 0
         parts = self.score.getElementsByClass(stream.Part)
@@ -203,7 +315,6 @@ class ScoreBase:
                 for voice in voices:
                     for note in voice:
                         for point in offsets:
-                            #print(note.offset, point)
                             if point >= note.offset and point < (note.offset + note.quarterLength):
                                 note_obj = NoteListElement(note, self.score.metadata, part.partName, part_number, note.quarterLength, self.url)
                                 pure_notes.append(note_obj)
@@ -212,6 +323,18 @@ class ScoreBase:
 
     # Allows for very specific note selection
     def note_list_single_part(self, part, measure_start, num_measures):
+        """
+        Creates a note list from a selected measure range within a single voice
+
+        Parameters
+        ----------
+        part : int
+            part number
+        measure_start : int
+            starting measure
+        num_measures : int
+            measures until end measure
+        """
         pure_notes = []
         part_selected = self.score.getElementsByClass(stream.Part)[part]
         measures = part_selected.getElementsByClass(stream.Measure)
@@ -236,6 +359,16 @@ class ScoreBase:
 
     # Allows for specific selection in terms of measures, but gets all parts/instruments
     def note_list_all_parts(self, measure_start, num_measures):
+        """
+        Creates a note list from a selected measure range over all voices
+
+        Parameters
+        ----------
+        measure_start : int
+            starting measure
+        num_measures : int
+            measures until end measure
+        """
         pure_notes = []
         parts = self.score.getElementsByClass(stream.Part)
         for part in parts:
@@ -261,8 +394,20 @@ class ScoreBase:
             pure_notes.append(note_obj)
         return pure_notes
 
-# An individual vector, not directly used by the user
+
 class VectorInterval:
+    """
+    An individual vector with information about the notes creating it
+
+    Attributes
+    ----------
+    vector : int or str
+        vector- in generic or semitones: is "Rest" if done between a note and a rest
+    note1 : NoteListElement
+        first note of interval pair
+    note2 : NoteListElement
+        list of notes constructed from score
+    """
     def __init__(self, vector, note1: NoteListElement, note2: NoteListElement):
         self.vector = vector
         self.note1 = note1
@@ -275,14 +420,40 @@ class VectorInterval:
             return "<VectorInterval: {}, First Note: {}, Second Note: {}>".format(self.vector, self.note1.note.nameWithOctave, self.note2.note.nameWithOctave)
 
 # Allows for selected "vectorizations" given a note list created from either ScoreBase or CorpusBase
+# Consider making this a Standalone method- an object seems slightly redundant/hard to justify
 class IntervalBase:
+    """
+    A list of VectorInterval objects created from a note list
+
+    Attributes
+    ----------
+    notes : list
+        note list gathered from either CorpusBase or ScoreBase's methods/attributes
+    generic_intervals : list
+        creates list of VectorInterval objects in terms of generic intervals
+    semitone_intervals : list
+        creates list of VectorInterval objects in terms of semitone intervals
+    """
     def __init__(self, notes):
+        """
+        Parameters
+        ----------
+        notes:
+            note list gathered from either CorpusBase or ScoreBase's methods/attributes
+        """
         self.notes = notes
         self.generic_intervals = self.vectorize_generic(self.notes)
         self.semitone_intervals = self.vectorize_semitone(self.notes)
 
     # Construct intervals in terms of semitone distances between notes
     def vectorize_semitone(self, notes):
+        """Creates list of VectorInterval objects in terms of semitone intervals
+
+        Parameters
+        ----------
+        notes:
+            (frequently self.notes): note list gathered from either CorpusBase or ScoreBase's methods/attributes
+        """
         vec = []
         for i in range(len(notes)-1):
             if notes[i].note.isRest or notes[i+1].note.isRest:
@@ -296,6 +467,13 @@ class IntervalBase:
 
     # Construct intervals in terms of generic distance between notes
     def vectorize_generic(self, notes):
+        """Creates list of VectorInterval objects in terms of generic intervals
+
+        Parameters
+        ----------
+        notes:
+            (frequently self.notes): note list gathered from either CorpusBase or ScoreBase's methods/attributes
+        """
         vec = []
         for i in range(len(notes)-1):
             if notes[i].note.isRest or notes[i+1].note.isRest:
@@ -309,6 +487,24 @@ class IntervalBase:
 
 # An individual match event- can be used for close matches as well
 class Match:
+    """
+    A pattern that has been deemed part of a match
+
+    Attributes
+    ----------
+    pattern : list
+        list of vectors in pattern
+    first_note : NoteListElement
+        first note in the soggetti creating the vector pattern
+    last_note : NoteListElement
+        last note in the soggetti creating the vector pattern
+    durations : list
+        list of durations of notes in soggetti creating the vector pattern
+    ema : str
+        standalone ema snippet for the pattern
+    ema_url : str
+        url to get mei for the pattern
+    """
     def __init__(self, pattern, first_note: NoteListElement, last_note: NoteListElement, durations):
         self.pattern = pattern
         self.first_note = first_note
@@ -330,12 +526,23 @@ class Match:
 # Object representing all the occurences of a pattern in a list of notes/vectors
 # User generally doesn't create this- it is done in the finding matches methods
 class PatternMatches:
+    """
+    A group of Match objects generated from a pattern
+
+    Attributes
+    ----------
+    pattern : list
+        pattern generating matches
+    matches : list
+        list of Match objects found to be matching the pattern
+    """
     def __init__(self, pattern, matches:list):
         self.pattern = pattern
-        # matches is a list of Match objects with the same pattern
         self.matches = matches
 
     def print_exact_matches(self):
+        """A facilitated way to display all the matches gathered by a find_exact_matches search
+        """
         print("Melodic interval/pattern " + str(self.pattern) + " occurs " + str(len(self.matches)) + " times:")
         for match in self.matches:
             print("In " + str(match.first_note.metadata.title) + " part " + str(match.first_note.part) + " beginning in measure " + str(match.first_note.note.measureNumber) +\
@@ -343,14 +550,31 @@ class PatternMatches:
         print("\n")
 
     def print_close_matches(self):
+        """A facilitated way to display all the matches gathered by a find_close_matches search
+        """
         print("Occurences of " + str(self.pattern) + " or similar:")
         for match in self.matches:
             print("Pattern " + str(match.pattern) + " appears in " + str(match.first_note.metadata.title) + " part " + str(match.first_note.part) + " beginning in measure " + str(match.first_note.note.measureNumber) +\
             " and ending in measure " + str(match.last_note.note.measureNumber) + ". Notes lengths: " + str(match.durations))
         print("Said pattern or similar appeared " + str(len(self.matches)) + " times.\n")
 
-# Standalone methods for match analysis
-def find_exact_matches(patterns_data, min_matches):
+# Potential redesign needed due to unstable nature of having user give over patterns_data
+# Potential fix is reincorporating into_patterns back into this method
+def find_exact_matches(patterns_data, min_matches=5):
+    """Takes in a series of vector patterns with data attached and finds exact matches
+
+    Parameters
+    ----------
+    patterns_data : return value from into_patterns
+        MUST be return value from into_patterns
+    min_matches : int, optional
+        Minimum number of matches needed to be deemed relevant, defaults to 5
+
+    Returns
+    -------
+    all_matches_list : list
+        A list of PatternMatches objects
+    """
     # A series of arrays are needed to keep track of various data associated with each pattern
     print("Finding exact matches...")
     patterns_nodup, patterns = [], []
@@ -379,6 +603,22 @@ def find_exact_matches(patterns_data, min_matches):
 
 # Finds matches based on a cumulative distance difference between two patterns
 def find_close_matches(patterns_data, min_matches, threshold):
+    """Takes in a series of vector patterns with data attached and finds close matches
+
+    Parameters
+    ----------
+    patterns_data : return value from into_patterns
+        MUST be return value from into_patterns
+    min_matches : int, optional
+        Minimum number of matches needed to be deemed relevant, defaults to 5
+    threshold : int
+        Cumulative variance allowed between vector patterns before they are deemed not similar
+
+    Returns
+    -------
+    all_matches_list : list
+        A list of PatternMatches objects
+    """
     # A series of arrays are needed to keep track of various data associated with each pattern
     print("Finding close matches...")
     patterns_nodup = []
@@ -407,6 +647,20 @@ def find_close_matches(patterns_data, min_matches, threshold):
 # Allows for the addition of non-moving-window pattern searching approaches
 # Needs to be called before any matches can be made
 def into_patterns(vectors_list, interval):
+    """Takes in a series of vector patterns with data attached and finds close matches
+
+    Parameters
+    ----------
+    vectors_list : list of vectorized lists
+        MUST be a list from calling generic_intervals or semitone_intervals on a VectorInterval object
+    interval : int
+        size of interval to be analyzed
+
+    Returns
+    -------
+    patterns_data : list of tuples
+        A list of vector patterns with additional information about notes attached
+    """
     pattern, patterns_data = [], []
     for vectors in vectors_list:
         for i in range(len(vectors)-interval):
@@ -426,15 +680,33 @@ def into_patterns(vectors_list, interval):
 
 # Helper for sort_matches
 def sortFunc(pattern):
+    """Helper function for sort_matches
+    """
     return len(pattern.matches)
 
 # Sorting based on the amount of matches each pattern has
 def sort_matches(matches_list):
+    """Sorts and returns a list of PatternMatch objects, ordering by size
+    """
     matches_list.sort(reverse=True, key=sortFunc)
     return matches_list
 
 # Generates a score from 0-1 based on how many patterns within a piece can be found in the other
 def similarity_score(notes1, notes2):
+    """Returns a score from 0-1 of the similarity between two note lists
+
+    Parameters
+    ----------
+    notes1 : list of NoteListElement objects
+        a note list from the CorpusBase or ScoreBase methods
+    notes2 : list of NoteListElement objects
+        a note list from the CorpusBase or ScoreBase methods
+
+    Returns
+    -------
+    final_score : int
+        a score of similarity from 0-1
+    """
     vectors1 = IntervalBase(notes1).generic_intervals
     vectors2 = IntervalBase(notes2).generic_intervals
     interval = 3
@@ -500,10 +772,22 @@ def similarity_score(notes1, notes2):
                         break
         interval += 1
         scores.append(score / (len(patterns_nodup2) + len(patterns_nodup1)))
-    return (scores[0] + scores[1] + scores[2] + scores[3]) / 4
+    final_score = (scores[0] + scores[1] + scores[2] + scores[3]) / 4
+    return final_score
 
 # Find all occurences of a specified pattern within a corpus
-def find_motif(pieces: CorpusBase, motif: list, generic: bool):
+def find_motif(pieces: CorpusBase, motif: list, generic: bool = True):
+    """Prints out all occurences of a specified motif
+
+    Parameters
+    ----------
+    pieces : CorpusBase
+        a CorpusBase object with all scores to be searched
+    motif : list
+        the motif in vectors (e.g. [-2,-2,2,-2,2])
+    generic : bool, optional
+        True to use generic vectors, False for semitone vectors- default is generic
+    """
     # Assemble into patterns
     vectors = IntervalBase(pieces.note_list)
     if generic:
@@ -521,6 +805,13 @@ def find_motif(pieces: CorpusBase, motif: list, generic: bool):
 
 # Given list of matches, write to csv in current working directory
 def export_to_csv(matches: list):
+    """Exports matches data to a csv in the current working directory
+
+    Parameters
+    ----------
+    matches : list
+        return value from either find_exact_matches or find_close_matches
+    """
     proceed = input("This method will create a csv file in your current working directory. Continue? (y/n): ").lower()
     csv_name = input("Enter a name for your csv file (.csv will be appended): ")
     csv_name += '.csv'
@@ -539,6 +830,13 @@ def export_to_csv(matches: list):
 # For more naive usage- allows for user interaction, has return value of list of matches
 # All features incorporated except non-whole piece selection
 def assisted_interface():
+    """Activates the assisted interface for more naive use
+
+    Returns
+    ----------
+    matches : list
+        list of PatternMatches based on the users various inputs
+    """
     print("You can use ctrl-c to quit exit at any time. If you proceed through the entire process, the matches array will be returned from this function")
     urls = []
     url = input("Enter a url, or 'done' when finished: ")
@@ -577,10 +875,11 @@ def assisted_interface():
                 item.print_exact_matches()
     return matches
 
-def sortMatches(match):
-    return match.first_note.offset
-
 def compare_durations(durations1, durations2, threshold):
+    """Used in classify_matches
+
+    works similarly to find_close_matches in terms of its comparison technique
+    """
     total = 0
     durations1_sum, durations2_sum = 0, 0
     for i in range(len(durations1)):
@@ -592,7 +891,35 @@ def compare_durations(durations1, durations2, threshold):
     else:
         return False
 
-def classify_matches(exact_matches: list, duration_threshold):
+def sortMatches(match):
+    """ Helper function for classify_matches
+    """
+    return match.first_note.offset
+
+
+def classify_matches(exact_matches: list, durations_threshold = 2):
+    """Classifies groups of matches into periodic entries, imitative duos, and fuga
+
+    Classifies through offset comparison of matching melodic patterns, prints out information gathered.
+    Reliably accurate results only guaranteed if exact_matches is generated from ONE piece.
+
+    Parameters
+    ----------
+    exact_matches : list
+        return value from find_exact_matches
+    durations_threshold : int, optional
+        maximum cumulative difference between two duration lists before they are deemed not similar, defaults to 2
+
+    Returns
+    -------
+    classified_tuple : tuple
+        classified_tuple[0] : list of lists of Match objects
+            list of periodic entries, which are lists of Match objects
+        classified_tuple[1] : list of lists of Match objects
+            list of imitative_duos, which are lists of Match objects
+        classified_tuple[0] : list of lists of Match objects
+            list of fuga, which are lists of Match objects
+    """
     # TO-DO: add in factoring for durations, narrow 80 beats window
     periodic_entries, im_duos, fuga = [], [], []
     for list_matches in exact_matches:
@@ -608,14 +935,23 @@ def classify_matches(exact_matches: list, duration_threshold):
             if offset_difs[i] > 64 or offset_difs[i + 1] > 64:
                 pass
             elif offset_difs[i] == offset_difs[i + 1] and offset_difs[i] == offset_difs[i + 2]:
-                periodic_entries.append((offset_difs_info[i][0], offset_difs_info[i][1], offset_difs_info[i + 1][0], offset_difs_info[i + 1][1], offset_difs_info[i + 2][0], offset_difs_info[i + 2][1]))
+                grouping = (offset_difs_info[i][0], offset_difs_info[i][1], offset_difs_info[i + 1][0], offset_difs_info[i + 1][1], offset_difs_info[i + 2][0], offset_difs_info[i + 2][1])
+                grouping = list(dict.fromkeys(grouping))
+                periodic_entries.append(grouping)
             elif offset_difs[i] == offset_difs[i + 1]:
-                periodic_entries.append((offset_difs_info[i][0], offset_difs_info[i][1], offset_difs_info[i + 1][0], offset_difs_info[i + 1][1]))
+                grouping = (offset_difs_info[i][0], offset_difs_info[i][1], offset_difs_info[i + 1][0], offset_difs_info[i + 1][1])
+                grouping = list(dict.fromkeys(grouping))
+                periodic_entries.append(grouping)
             elif offset_difs[i] == offset_difs[i + 2]:
-                im_duos.append((offset_difs_info[i][0], offset_difs_info[i][1], offset_difs_info[i + 2][0], offset_difs_info[i + 2][1]))
+                grouping = (offset_difs_info[i][0], offset_difs_info[i][1], offset_difs_info[i + 2][0], offset_difs_info[i + 2][1])
+                grouping = list(dict.fromkeys(grouping))
+                im_duos.append(grouping)
             else:
-                fuga.append((offset_difs_info[i][0], offset_difs_info[i][1], offset_difs_info[i + 1][0], offset_difs_info[i + 1][1]))
+                grouping = (offset_difs_info[i][0], offset_difs_info[i][1], offset_difs_info[i + 1][0], offset_difs_info[i + 1][1])
+                grouping = list(dict.fromkeys(grouping))
+                fuga.append(grouping)
             i += 1
+
     for entry in periodic_entries:
         print("Periodic Entry:")
         a = "Pattern: " + str(entry[0].pattern) + ", Locations in entry: "
@@ -624,14 +960,15 @@ def classify_matches(exact_matches: list, duration_threshold):
         print(a)
     for duo in im_duos:
         print("Imitative Duo:")
-        a = "Pattern: " + str(entry[0].pattern) + ", Locations in entry: "
+        a = "Pattern: " + str(duo[0].pattern) + ", Locations in entry: "
         for c in duo:
             a += "\n- Measure " + str(c.first_note.note.measureNumber) + " in voice " + str(c.first_note.partNumber)
         print(a)
     for f in fuga:
         print("Fuga:")
-        a = "Pattern: " + str(entry[0].pattern) + ", Locations in entry: "
+        a = "Pattern: " + str(f[0].pattern) + ", Locations in entry: "
         for d in f:
             a += "\n- Measure " + str(d.first_note.note.measureNumber) + " in voice " + str(d.first_note.partNumber)
         print(a)
-    return ((periodic_entries, im_duos, fuga))
+    classified_tuple = (periodic_entries, im_duos, fuga)
+    return classified_tuple
